from __future__ import unicode_literals

INITIAL_SPACE = "     "
fmt_package_headers = (
    INITIAL_SPACE + "%(channel_path)-30s | %(name)-40s | %(version)8s | %(family)-12s "
    "| %(build_number)-10s | %(license)-15s | %(subdirs)-15s"
)

fmt_package_simple_headers = (
    INITIAL_SPACE + "%(name)-40s | %(file_count)-12s | %(download_count)-12s "
    "| %(license)-15s"
)
fmt_channel_headers = (
    INITIAL_SPACE
    + "%(channel_path)-30s | %(privacy)-10s | %(owners)15s | %(artifact_count)-12s "
    "| %(download_count)-11s | %(subchannel_count)-14s |  %(mirror_count)-9s | %(description)-30s"
)
fmt_mirror_header_spacer = {
    "id": "-" * 36,
    "name": "-" * 15,
    "type": "-" * 8,
    "mode": "-" * 10,
    "state": "-" * 10,
    "source_root": "-" * 50,
    "proxy": "-" * 40,
    "last_run_at": "-" * 30,
    "updated_at": "-" * 30,
}
fmt_mirror_headers = (
    INITIAL_SPACE + "%(id)-36s | %(name)-15s | %(type)8s | %(mode)-10s "
    "| %(state)-10s | %(source_root)-50s | %(proxy)-40s | %(last_run_at)-30s | %(updated_at)-30s"
)

fmt_all_mirror_header_spacer = {
    "id": "-" * 36,
    "name": "-" * 15,
    "source_root": "-" * 45,
    "type": "-" * 18,
    "privacy": "-" * 14,
    "channel": "-" * 18,
    "mode": "-" * 10,
    "state": "-" * 10,
    "last_run_at": "-" * 34,
    "created_at": "-" * 34,
    "updated_at": "-" * 34,
}

fmt_all_mirror_headers = (
    INITIAL_SPACE
    + "%(id)-36s | %(name)-15s | %(source_root)45s | %(type)-18s | %(channel)-18s  | %(privacy)-14s  "
    "| %(mode)-10s | %(state)-10s | %(last_run_at)-34s | %(created_at)-34s | %(updated_at)-34s"
)

fmt_sync_state_header_spacer = {
    "active": "-" * 10,
    "failed": "-" * 10,
    "passive": "-" * 10,
    "removed": "-" * 10,
    "last_ckey": "-" * 60,
    "last_exception": "-" * 18,
    "total_steps_count": "-" * 18,
    "current_step_number": "-" * 20,
    "current_step_percentage": "-" * 25,
    "current_step_description": "-" * 34,
}
fmt_sync_state_headers = (
    INITIAL_SPACE
    + "%(active)-10s | %(failed)-10s | %(passive)-10s | %(removed)-10s | %(last_ckey)-60s  | %(last_exception)-18s  "
    "| %(total_steps_count)-18s | %(current_step_number)-20s | %(current_step_percentage)-25s | %(current_step_description)-34s"
)

fmt_sync_state_mirror_filtered_header_spacer = {
    "cve": "-" * 18,
    "date": "-" * 18,
    "license": "-" * 18,
    "only_specs": "-" * 18,
    "exclude_specs": "-" * 18,
    "duplicated_legacy_conda": "-" * 18,
}
fmt_sync_state_mirror_filtered_header = (
    INITIAL_SPACE
    + "%(cve)-18s | %(date)-18s | %(license)-18s | %(only_specs)-18s | %(exclude_specs)-18s  | %(duplicated_legacy_conda)-18s"
)

fmt_sync_state_channel_filtered_header_spacer = {
    "license": "-" * 18,
    "exclude_specs": "-" * 18,
}
fmt_sync_state_channel_filtered_header = (
    INITIAL_SPACE + "%(license)-18s | %(exclude_specs)-18s"
)


class PackagesFormatter:
    def __init__(self, log):
        self.log = log

    @staticmethod
    def format_package_header(simple=False):
        if simple:
            package_header = {
                "name": "Name",
                "file_count": "# of files",
                "download_count": "# of download",
                "license": "License",
            }
            fmt = fmt_package_simple_headers
        else:
            package_header = {
                "channel_path": "Channel",
                "name": "Name",
                "family": "Family",
                "version": "Version",
                "subdirs": "Platforms",
                "license": "License",
                "build_number": "Build",
            }
            fmt = fmt_package_headers
        return fmt % package_header

    def log_format_package_header(self, simple=False):
        self.log.info(self.format_package_header(simple=simple))

    @staticmethod
    def format_package(package, simple=False):
        package = package.copy()
        package.update(package["metadata"])

        if package["subchannel"]:
            package["channel_path"] = "%s/%s" % (
                package["channel"],
                package["subchannel"],
            )
        else:
            package["channel_path"] = package["channel"]

        if "subdirs" not in package:
            package["subdirs"] = []

        if "build_number" not in package:
            package["build_number"] = ""

        if "license" not in package:
            package["license"] = ""

        package["full_name"] = "%s::%s" % (package["channel_path"], package["name"])
        package["subdirs"] = ", ".join(
            str(x) for x in package["subdirs"] if x is not None
        )

        if simple:
            fmt = fmt_package_simple_headers
        else:
            fmt = fmt_package_headers

        return fmt % package

    def log_format_package(self, package, simple=False):
        self.log.info(self.format_package(package, simple=simple))

    @staticmethod
    def format_channel_header():
        package_header = {
            "channel_path": "Channel",
            "privacy": "Privacy",
            "owners": "Owners",
            "artifact_count": "# Artifacts",
            "download_count": "# Downloads",
            "subchannel_count": "# Subchannels",
            "mirror_count": "# Mirrors",
            "description": "Description",
        }
        return fmt_channel_headers % package_header

    def log_format_channel_header(self):
        self.log.info(self.format_channel_header())

    @staticmethod
    def format_channel(channel):
        channel = channel.copy()

        if channel["parent"]:
            channel["channel_path"] = "%s/%s" % (channel["parent"], channel["name"])
        else:
            channel["channel_path"] = channel["name"]

        channel["owners"] = ", ".join(
            str(x) for x in channel["owners"] if x is not None
        )
        return fmt_channel_headers % channel

    def log_format_channel(self, channel):
        self.log.info(self.format_channel(channel))

    def format(self, packages, metadata, simple=False):
        self.log_format_package_header(simple=simple)

        if simple:
            package_header = {
                "name": "-" * 40,
                "file_count": "-" * 8,
                "download_count": "-" * 12,
                "license": "-" * 15,
            }

            self.log.info(fmt_package_simple_headers % package_header)
        else:
            package_header = {
                "channel_path": "-" * 30,
                "name": "-" * 40,
                "version": "-" * 8,
                "family": "-" * 12,
                "subdirs": "-" * 15,
                "license": "-" * 15,
                "build_number": "-" * 10,
            }

            self.log.info(fmt_package_headers % package_header)

        for package in packages:
            self.log_format_package(package, simple=simple)

        if packages:
            end_set = len(packages) + metadata["offset"]
            self.log.info(
                "\n%s%i packages found." % (INITIAL_SPACE, metadata["total_count"])
            )
            self.log.info(
                "%sVisualizing %i-%i interval."
                % (INITIAL_SPACE, len(packages), end_set)
            )
        else:
            self.log.info("No packages found")

        self.log.info("")


class MirrorFormatter:
    keymap = {
        "created_at": "created at",
        "updated_at": "Updated at",
        "last_run_at": "Last run at",
    }

    @staticmethod
    def format_detail(mirror):
        mirror = {key.replace("mirror_", ""): val for key, val in mirror.items()}
        keymap = {
            "created_at": "created at",
            "updated_at": "Updated at",
            "last_run_at": "Last run at",
        }
        mirror_ = dict(mirror)
        resp = [INITIAL_SPACE + "Mirror Details:", INITIAL_SPACE + "---------------"]

        fields = [
            "id",
            "name",
            "type",
            "mode",
            "state",
            "source_root",
            "last_run_at",
            "updated_at",
            "created",
            "cron",
            "proxy",
            "filters",
        ]
        for key in fields:
            label = keymap.get(key, key.replace("_", " ")).capitalize()
            value = mirror_.get(key, "")
            if key == "filters":
                resp.append("%s%s:" % (INITIAL_SPACE, label))
                if value:
                    for filter_key in sorted(value.keys()):
                        resp.append(
                            "%s   %s| %s"
                            % (INITIAL_SPACE, filter_key.ljust(25), value[filter_key])
                        )
            else:
                resp.append("%s%s: %s" % (INITIAL_SPACE, label, value))

        return "\n".join(resp)

    @classmethod
    def format_list_headers(cls):
        mirror_headers = {k: k.capitalize() for k in fmt_mirror_header_spacer}
        mirror_headers.update(cls.keymap)
        return fmt_mirror_headers % mirror_headers

    @staticmethod
    def format_list_item(mirror):
        mirror = {
            key.replace("mirror_", ""): (val if val is not None else "")
            for key, val in mirror.items()
        }
        return fmt_mirror_headers % mirror

    @staticmethod
    def format_list(mirrors):
        lines = []
        lines.append(MirrorFormatter.format_list_headers())
        lines.append(fmt_mirror_headers % fmt_mirror_header_spacer)

        for mirror in mirrors:
            lines.append(MirrorFormatter.format_list_item(mirror))

        return "\n".join(lines)


class AllMirrorFormatter:
    keymap = {
        "source_root": "root source",
        "last_run_at": "last run at",
        "created_at": "created at",
        "updated_at": "updated at",
    }

    @staticmethod
    def format_list_all(mirrors):
        lines = []
        lines.append(AllMirrorFormatter.format_list_all_headers())
        lines.append(fmt_all_mirror_headers % fmt_all_mirror_header_spacer)

        for mirror in mirrors:
            lines.append(AllMirrorFormatter.format_list_item(mirror))

        return "\n".join(lines)

    @classmethod
    def format_list_all_headers(cls):
        mirror_headers = {k: k.capitalize() for k in fmt_all_mirror_header_spacer}
        mirror_headers.update(cls.keymap)
        return fmt_all_mirror_headers % mirror_headers

    @staticmethod
    def format_list_item(mirror):
        mirror.pop("sync_state", None)

        channel = mirror["channel"]
        mirror.pop("channel", None)
        channel["channel"] = channel.pop("name")
        mirror = {**mirror, **channel}

        return fmt_all_mirror_headers % mirror


class SyncStateFormatter:
    keymap = {
        "last_ckey": "Last ckey",
        "last_exception": "Last exception",
        "total_steps_count": "Total steps count",
        "current_step_number": "Current step number",
        "current_step_percentage": "Current step percentage",
        "current_step_description": "Current step description",
    }
    keymap_mirror = {
        "only_specs": "Only specs",
        "exclude_specs": "exclude specs",
        "duplicated_legacy_conda": "Duplicated legacy conda",
    }
    keymap_channel = {"exclude_specs": "Exclude specs"}

    @staticmethod
    def format_sync_state(data, mirror_name):
        mirror = None
        for m in data:
            if m["name"] == mirror_name:
                mirror = m

        if not mirror:
            lines = []
            lines.append("No Mirror named %s" % mirror_name)
            return "\n".join(lines)

        sync_state = mirror["sync_state"]

        lines = []

        lines.append(INITIAL_SPACE + "SYNC STATE:")
        lines.append(SyncStateFormatter.format_sync_state_headers())
        lines.append(fmt_sync_state_headers % fmt_sync_state_header_spacer)
        lines.append(SyncStateFormatter.format_sync_state_general(sync_state))
        lines.append("")

        lines.append(INITIAL_SPACE + "FILTERED PACKAGES BY MIRROR:")
        lines.append(SyncStateFormatter.format_filtered_mirror_headers())
        lines.append(
            fmt_sync_state_mirror_filtered_header
            % fmt_sync_state_mirror_filtered_header_spacer
        )
        lines.append(
            SyncStateFormatter.format_sync_state_filtered_mirror(sync_state["count"])
        )
        lines.append("")

        lines.append(INITIAL_SPACE + "FILTERED PACKAGES BY CHANNEL:")
        lines.append(SyncStateFormatter.format_filtered_channel_headers())
        lines.append(
            fmt_sync_state_channel_filtered_header
            % fmt_sync_state_channel_filtered_header_spacer
        )
        lines.append(
            SyncStateFormatter.format_sync_state_filtered_channel(sync_state["count"])
        )
        lines.append("")

        return "\n".join(lines)

    @classmethod
    def format_sync_state_headers(cls):
        sync_state_headers = {k: k.capitalize() for k in fmt_sync_state_header_spacer}
        sync_state_headers.update(cls.keymap)
        return fmt_sync_state_headers % sync_state_headers

    @staticmethod
    def format_sync_state_general(sync_state):
        sync_state = sync_state.copy()
        count = sync_state["count"]
        sync_state.pop("count", None)
        sync_state.update(count)
        sync_state.pop("mirror_filtered", None)
        sync_state.pop("channel_filtered", None)

        return fmt_sync_state_headers % sync_state

    @classmethod
    def format_filtered_mirror_headers(cls):
        sync_state_mirror_filtered_header = {
            k: k.capitalize() for k in fmt_sync_state_mirror_filtered_header_spacer
        }
        sync_state_mirror_filtered_header.update(cls.keymap_mirror)
        return fmt_sync_state_mirror_filtered_header % sync_state_mirror_filtered_header

    @staticmethod
    def format_sync_state_filtered_mirror(sync_state):
        mirror_filtered = sync_state["mirror_filtered"]
        return fmt_sync_state_mirror_filtered_header % mirror_filtered

    @classmethod
    def format_filtered_channel_headers(cls):
        sync_state_channel_filtered_header = {
            k: k.capitalize() for k in fmt_sync_state_channel_filtered_header_spacer
        }
        sync_state_channel_filtered_header.update(cls.keymap_channel)
        return (
            fmt_sync_state_channel_filtered_header % sync_state_channel_filtered_header
        )

    @staticmethod
    def format_sync_state_filtered_channel(sync_state):
        channel_filtered = sync_state["channel_filtered"]
        return fmt_sync_state_channel_filtered_header % channel_filtered


class FormatterBase:
    entity = ""
    fmt_header_spacer = {}
    fmt_headers = INITIAL_SPACE + ""

    keymap = {
        "date_added": "Published",
        "updated_at": "Updated at",
        "last_run_at": "Last run at",
    }

    fields = []

    @classmethod
    def format_detail(cls, item):
        item_ = cls.normalize_item(item)
        resp = [
            INITIAL_SPACE + "%s Details:" % cls.entity,
            INITIAL_SPACE + "---------------",
        ]

        for key in cls.fields:
            label = cls.keymap.get(key, key).capitalize()
            value = nested_get(item_, key, "")
            resp.append("%s%s: %s" % (INITIAL_SPACE, label, value))

        return "\n".join(resp)

    @classmethod
    def format_list_headers(cls):
        list_headers = {k: k.capitalize() for k in cls.fmt_header_spacer}
        list_headers.update(cls.keymap)
        return cls.fmt_headers % list_headers

    @classmethod
    def format_list_item(cls, item):
        item_ = cls.normalize_item(item)
        return cls.fmt_headers % item_

    @staticmethod
    def normalize_item(item):
        raise NotImplementedError

    @classmethod
    def format_list(cls, items):
        lines = []
        lines.append(cls.format_list_headers())
        lines.append(cls.fmt_headers % cls.fmt_header_spacer)

        for item in items:
            lines.append(cls.format_list_item(item))

        return "\n".join(lines)


class CVEFormatter(FormatterBase):
    entity = "CVE"
    fmt_header_spacer = {
        "id": "-" * 14,
        "score": "-" * 6,
        "score_type": "-" * 10,
        "curated": "-" * 8,
        "packages_count": "-" * 10,
        "description": "-" * 100,
    }
    fmt_headers = (
        INITIAL_SPACE
        + "%(id)-14s | %(score)6s | %(score_type)10s | %(curated)8s | %(packages_count)10s | %(description)-50s"
    )

    keymap = {
        "id": "CVE ID",
        "score_type": "Type",
        "packages_count": "# Packages",
        "cvssv2.cvssV2.accessComplexity": "CVSS V2 Access Complexity",
        "cvssv2.cvssV2.accessVector": "CVSS V2 Access Vector",
        "cvssv2.cvssV2.authentication": "CVSS V2 Authentication",
        "cvssv2.cvssV2.availabilityImpact": "CVSS V2 Availablity Impact",
        "cvssv2.cvssV2.confidentialityImpact": "CVSS V2 Confidentiality Impact",
        "cvssv2.cvssV2.integrityImpact": "CVSS V2 Integrity Impact",
        "cvssv2.cvssV2.baseScore": "CVSS V2 Base Score",
        "cvssv2.severity": "CVSS V2 Severity",
        "cvssv3.cvssV3.attackComplexity": "CVSS V3 Attack Complexity",
        "cvssv3.cvssV3.attackVector": "CVSS V3 Attack Vector",
        "cvssv3.cvssV3.availabilityImpact": "CVSS V3 Availability Impact",
        "cvssv3.cvssV3.baseScore": "CVSS V3 Base Score",
        "cvssv3.cvssV3.baseSeverity": "CVSS V3 Base Severity",
        "cvssv3.cvssV3.confidentialityImpact": "CVSS V3 Confidentiality Impact",
        "cvssv3.cvssV3.integrityImpact": "CVSS V3 Integrity Impact",
        "cvssv3.cvssV3.privilegesRequired": "CVSS V3 Privileges Required",
        "cvssv3.cvssV3.scope": "CVSS V3 Scope",
        "cvssv3.cvssV3.userInteraction": "CVSS V3 User Interaction",
    }

    fields = [
        "id",
        "curated",
        "score",
        "score_type",
        "description",
        "cvssv2.cvssV2.accessComplexity",
        "cvssv2.cvssV2.accessVector",
        "cvssv2.cvssV2.authentication",
        "cvssv2.cvssV2.availabilityImpact",
        "cvssv2.cvssV2.confidentialityImpact",
        "cvssv2.cvssV2.integrityImpact",
        "cvssv2.cvssV2.baseScore",
        "cvssv2.severity",
        "cvssv3.cvssV3.attackComplexity",
        "cvssv3.cvssV3.attackVector",
        "cvssv3.cvssV3.availabilityImpact",
        "cvssv3.cvssV3.baseScore",
        "cvssv3.cvssV3.baseSeverity",
        "cvssv3.cvssV3.confidentialityImpact",
        "cvssv3.cvssV3.integrityImpact",
        "cvssv3.cvssV3.privilegesRequired",
        "cvssv3.cvssV3.scope",
        "cvssv3.cvssV3.userInteraction",
        "published_at",
        "packages_count",
    ]

    @staticmethod
    def normalize_item(item):
        item_ = {key: val for key, val in item.items()}
        score = item_.get("cvssv3_score")
        if score:
            item_["score"] = score
            item_["score_type"] = "CVSS3"
        else:
            item_["score"] = item_.get("cvssv2_score")
            item_["score_type"] = "CVSS2"

        def fmt_package(pack):
            try:
                if "subdir" not in pack:
                    pack["subdir"] = ""
                return "{subdir}/{name}-{version} (sha258: {sha256})".format(**pack)
            except TypeError:
                return pack

        packages = [fmt_package(pack) for pack in item_.get("packages", [])]
        item_["packages"] = ", ".join(packages)
        return item_


class CVEFilesFormatter(FormatterBase):
    entity = "CVE Files"
    fmt_header_spacer = {
        "channel": "-" * 40,
        "artifact_family": "-" * 10,
        "cve_status": "-" * 10,
        "common_name": "-" * 18,
        "ckey": "-" * 32,
    }
    fmt_headers = (
        INITIAL_SPACE
        + "%(channel)-40s | %(artifact_family)-10s | %(cve_status)-10s | %(common_name)-18s | %(ckey)-32s"
    )

    keymap = {
        "cve_status": "Status",
        "artifact_family": "Family",
        "common_name": "Name",
        "ckey": "Path",
    }

    @staticmethod
    def normalize_item(item):
        item_ = {key: val for key, val in item.items()}
        item_["channel"] = (
            "%s/%s" % (item_["parent_channel_name"], item_["channel_name"])
            if item_["parent_channel_name"]
            else item_["channel_name"]
        )
        return item_


class PackageFilesFormatter(FormatterBase):
    entity = "Package Files"
    fmt_header_spacer = {
        "ckey": "-" * 60,
        "version": "-" * 10,
        "platform": "-" * 10,
    }
    fmt_headers = INITIAL_SPACE + "%(ckey)-60s | %(version)-10s | %(platform)-10s"

    keymap = {
        "ckey": "Path",
    }

    @staticmethod
    def normalize_item(item):
        return item


class PackageFilesFormatterWithCVE(FormatterBase):
    entity = "Package Files"
    fmt_header_spacer = {
        "ckey": "-" * 60,
        "cve_score": "-" * 9,
        "cve_status": "-" * 10,
        "version": "-" * 10,
        "platform": "-" * 10,
    }
    fmt_headers = (
        INITIAL_SPACE
        + "%(ckey)-60s | %(cve_score)-9s | %(cve_status)-10s | %(version)-10s | %(platform)-10s"
    )

    keymap = {"ckey": "Path", "cve_score": "CVE Score", "cve_status": "CVE Status"}

    @staticmethod
    def normalize_item(item):
        item_ = {key: val for key, val in item.items()}
        if item_["cve_status"] is None:
            item_["cve_status"] = "n/a"
        if item_["cve_score"] is None:
            item_["cve_score"] = "n/a"
        return item_


class HistoryFormatter(FormatterBase):
    entity = "Events"
    fmt_header_spacer = {
        "event_id": "-" * 36,
        "event_type": "-" * 20,
        "created": "-" * 32,
        "data_summary": "-" * 100,
    }
    fmt_headers = (
        INITIAL_SPACE
        + "%(event_id)-36s | %(event_type)20s | %(created)32s | %(data_summary)-50s"
    )

    keymap = {"event_id": "id", "event_type": "Type", "data_summary": "Summary"}

    fields = ["created", "data", "event_id", "event_type", "meta", "data_summary"]
    short_summary_keys = {"ckey", "artifact_family"}

    @classmethod
    def format_list_item(cls, item, short_summary=True):
        item_ = cls.normalize_item(item, short_summary)
        return cls.fmt_headers % item_

    @classmethod
    def normalize_item(cls, item, short_summary=True):
        item_ = {key: val for key, val in item.items()}
        if short_summary:
            item_["data_summary"] = "; ".join(
                [
                    "%s: %s" % (k, v)
                    for k, v in item["data"].items()
                    if k in cls.short_summary_keys
                ]
            )
        else:
            item_["data_summary"] = "; ".join(
                ["%s: %s" % (k, v) for k, v in item["data"].items()]
            )
        return item_

    @classmethod
    def format_list(cls, items, short_summary=True):
        lines = []
        lines.append(cls.format_list_headers())
        lines.append(cls.fmt_headers % cls.fmt_header_spacer)

        for item in items:
            lines.append(cls.format_list_item(item, short_summary))

        return "\n".join(lines)


class SettingsFormatter(FormatterBase):
    entity = "Admin Settings"
    fmt_header_spacer = {
        "key": "-" * 36,
        "value": "-" * 12,
    }
    fmt_headers = INITIAL_SPACE + "%(key)-36s | %(value)-12s"

    keymap = {"key": "Setting Name", "value": "Value"}
    fields = ["key", "value"]

    @classmethod
    def normalize_item(cls, item, short_summary=True):
        return item

    @classmethod
    def format_object_as_list(cls, obj):
        return cls.format_list([{"key": k, "value": v} for k, v in obj.items()])


def format_packages(packages, meta, logger):
    formatter = PackagesFormatter(logger)
    formatter.format(packages, meta)
    return formatter


def nested_get(item, path, default=None):
    """Returns the value of the nested dictionary, using path.

    Example:
        nested_get({"a":{"b": 42}}, "a.b") == 42
    """
    result = item
    for key in path.split("."):
        if key not in result:
            return default
        result = result[key]
    return result


def comma_string_to_list(s):
    """Returns a list of strings from the string, splitting by comma and removing trailing empty
        characters.

    Example:
        comma_string_to_list("flask, numpy ") == ["flask", "numpy"]
    """
    return [item.strip() for item in s.split(",") if item.strip()]



from unittest.mock import Mock, patch

from repo_cli.commands import wizard
from repo_cli.tests.unit.test_utils.fixtures import fixtures as all_channels

mock_self = Mock()
mock_self.log.warning = Mock()
w = wizard.SubCommand(mock_self)


def test_purge_channels():
    expected_result = ["test"]

    input_value = "test c2 b1"
    channels = w.purge_channels(input_value, all_channels)

    assert channels == expected_result


@patch("builtins.input", side_effect=["c2", "test"])
def test_no_empty_defaults(monkeypatch):
    expected_result = ["test"]

    channels = w.get_default_channels(all_channels)
    # Assert that it failed when called with invalid channel 'c2'
    mock_self.log.warning.assert_called_with(
        "At least one channel should be added to default channels! Try again..."
    )
    assert channels == expected_result



"""
Utilities functions for Anaconda repository command line manager
"""
from __future__ import print_function, unicode_literals

import logging
import sys
from logging.handlers import RotatingFileHandler
from os import getenv, makedirs, path
from os.path import exists, isfile, join

import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from six import PY2

from .config import USER_LOGDIR


def file_or_token(value):
    """
    If value is a file path and the file exists its contents are stripped and returned,
    otherwise value is returned.
    """
    if isfile(value):
        with open(value) as fd:
            return fd.read().strip()

    if any(char in value for char in "/\\."):
        # This chars will never be in a token value, but may be in a path
        # The error message will be handled by the parser
        raise ValueError()

    return value


def _custom_excepthook(logger, show_traceback=False):
    def excepthook(exc_type, exc_value, exc_traceback):
        if issubclass(exc_type, KeyboardInterrupt):
            logger.error("execution interrupted")
            return

        if show_traceback:
            logger.error("", exc_info=(exc_type, exc_value, exc_traceback))
        else:
            logger.error("%s", exc_value)

    return excepthook


def get_ssl_verify_option(config, insecure, logger):
    """
    Tries to determine if SSL certificate verification is
    needed and if needed then the follows priorities below

    1. If `verify_ssl` is given true then it is highest priority,
    2. Then `SSL_NO_VERIFY` comes next if environment variable is set,
    3. Else if `ssl_verify` is set in config and if either string or boolean,
    4. Finally we should verify SSL certificates by default.
    """
    # Give priority to --insecure option if provided

    if insecure and insecure is not None:
        return False

    # Support environment setting like in conda
    insecure = getenv("SSL_NO_VERIFY")
    if insecure and insecure.isdigit() and int(insecure) == 1:
        logger.warning("SSL_NO_VERIFY is set please make sure to unset it.")
        return False

    # And least priority is given to setting based on configuration
    config_ssl_verify = config.get("ssl_verify")
    if isinstance(config_ssl_verify, (bool, str)):
        if isinstance(config_ssl_verify, bool) and not config_ssl_verify:
            logger.warning(
                '"ssl_verify" is set to false please consider to turn it on.'
            )

        if not path.exists(config_ssl_verify):
            logger.warning(
                'Specified "ssl_verify=%s" certificate path does not exist.'
                % config_ssl_verify
            )

        return config_ssl_verify

    return True


class ConsoleFormatter(logging.Formatter):
    def format(self, record):
        fmt = (
            "%(message)s"
            if record.levelno == logging.INFO
            else "[%(levelname)s] %(message)s"
        )
        if PY2:
            self._fmt = fmt
        else:
            self._style._fmt = fmt
        return super(ConsoleFormatter, self).format(record)


def _setup_logging(
    logger, log_level=logging.INFO, show_traceback=False, disable_ssl_warnings=False
):
    logger.setLevel(logging.DEBUG)

    if not exists(USER_LOGDIR):
        makedirs(USER_LOGDIR)

    log_file = join(USER_LOGDIR, "cli.log")

    file_handler = RotatingFileHandler(
        log_file, maxBytes=10 * (1024**2), backupCount=5
    )
    file_handler.setLevel(logging.DEBUG)

    console_handler = logging.StreamHandler(stream=sys.stdout)
    console_handler.setLevel(log_level)

    console_handler.setFormatter(ConsoleFormatter())
    file_handler.setFormatter(
        logging.Formatter("%(asctime)s %(levelname)-8s %(name)-15s %(message)s")
    )

    logger.addHandler(console_handler)
    logger.addHandler(file_handler)

    sys.excepthook = _custom_excepthook(logger, show_traceback=show_traceback)

    if disable_ssl_warnings:
        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)



"""
Manage your Anaconda repository channels.
"""

from __future__ import print_function, unicode_literals

import argparse

from .. import errors
from ..utils.format import (
    AllMirrorFormatter,
    MirrorFormatter,
    SyncStateFormatter,
    comma_string_to_list,
)
from ..utils.validators import (
    check_cve_score,
    check_date_format,
    check_proxy_url,
    check_url,
)
from .base import SubCommandBase

DEFAULT_MIRROR_CRON = "0 0 * * *"


class SubCommand(SubCommandBase):
    name = "mirror"

    def main(self):
        self.log.info("")
        args = self.args
        if (
            args.create or args.update or args.delete or args.show
        ) and not args.channel:
            msg = (
                "Channel name not specified. Please use -c or --channel to specify your channel.\n"
                "Use --help for help."
            )
            self.log.info(msg)
            raise errors.RepoCLIError(msg)
        if args.create:
            self.create_mirror(
                args.channel,
                args.source,
                args.create,
                args.mode,
                args.type,
                args.licenses,
                args.subdirs,
                args.projects,
                args.only_spec,
                args.exclude_spec,
                args.include_spec,
                args.cve_score,
                args.exclude_uncurated_cve_packages,
                args.date_from,
                args.date_to,
                args.cron,
                args.run_now,
                args.proxy,
            )
        elif args.update:
            self.update_mirror(
                args.channel,
                args.source,
                args.update,
                args.name,
                args.mode,
                args.type,
                args.licenses,
                args.subdirs,
                args.projects,
                args.only_spec,
                args.exclude_spec,
                args.include_spec,
                args.cve_score,
                args.exclude_uncurated_cve_packages,
                args.date_from,
                args.date_to,
                args.cron,
                args.run_now,
                args.proxy,
            )
        elif args.delete:
            self.delete(args.channel, args.delete)
        elif args.list:
            self.show_list(args.list)
        elif args.list_all:
            self.show_all_list()
        elif args.sync_state:
            self.sync_state(args.sync_state)
        elif args.show:
            self.show(args.channel, args.show)
        else:
            raise NotImplementedError()

    # flake8: noqa: C901
    def create_mirror(
        self,
        channel,
        source,
        name,
        mode,
        type_,
        filter_licenses,
        filter_subdirs,
        filter_projects,
        filter_only_specs,
        filter_exclude_specs,
        filter_include_specs,
        cve_score,
        exclude_uncurrated_cve_packages,
        filter_date_from,
        filter_date_to,
        cron,
        run_now,
        proxy,
    ):
        filters = {}
        if filter_subdirs:
            filters["subdirs"] = comma_string_to_list(filter_subdirs)
        if filter_projects:
            filters["projects"] = comma_string_to_list(filter_projects)
        if filter_licenses:
            filters["include_licenses"] = []
            filters["exclude_licenses"] = []
            for license in comma_string_to_list(filter_licenses):
                group = "include_licenses"
                if license.startswith("-"):
                    group = "exclude_licenses"
                    license = license[1:]
                if license.startswith("+"):
                    license = license[1:]
                filters[group].append(license)
        if filter_only_specs:
            filters["only_specs"] = filter_only_specs
        if filter_exclude_specs:
            filters["exclude_specs"] = filter_exclude_specs
        if filter_include_specs:
            filters["include_specs"] = filter_include_specs
        if filter_date_from:
            filters["date_from"] = filter_date_from
        if filter_date_to:
            filters["date_to"] = filter_date_to
        if cve_score is not None:
            filters["cve_score_threshold"] = cve_score
        if exclude_uncurrated_cve_packages:
            filters["exclude_non_curated_cve"] = exclude_uncurrated_cve_packages

        self.validate_mirror_filters(filters)

        self.api.create_mirror(
            channel, source, name, mode, filters, type_, cron, run_now, proxy
        )
        self.log.info("Mirror %s successfully created on channel %s", name, channel)

    def update_mirror(
        self,
        channel,
        source,
        name,
        new_name,
        mode,
        type_,
        filter_licenses,
        filter_subdirs,
        filter_projects,
        filter_only_specs,
        filter_exclude_specs,
        filter_include_specs,
        cve_score,
        exclude_uncurrated_cve_packages,
        filter_date_from,
        filter_date_to,
        cron,
        run_now,
        proxy,
    ):
        mirror = self._get_by_name(channel, name)
        if not mirror:
            self.log.info("Mirror not found")
            return

        filters = mirror["filters"]

        if filter_subdirs is not None:
            filters["subdirs"] = comma_string_to_list(filter_subdirs)
        if filter_projects is not None:
            filters["projects"] = comma_string_to_list(filter_projects)
        if filter_licenses is not None:
            filters["include_licenses"] = []
            filters["exclude_licenses"] = []
            for license in comma_string_to_list(filter_licenses):
                group = "include_licenses"
                if license.startswith("-"):
                    group = "exclude_licenses"
                    license = license[1:]
                if license.startswith("+"):
                    license = license[1:]
                filters[group].append(license)
        if filter_only_specs is not None:
            filters["only_specs"] = filter_only_specs
        if filter_exclude_specs is not None:
            filters["exclude_specs"] = filter_exclude_specs
        if filter_include_specs is not None:
            filters["include_specs"] = filter_include_specs
        if filter_date_from is not None:
            filters["date_from"] = filter_date_from
        if filter_date_to is not None:
            filters["date_to"] = filter_date_to
        if cve_score is not None:
            filters["cve_score_threshold"] = cve_score
        if exclude_uncurrated_cve_packages is not None:
            filters["exclude_non_curated_cve"] = exclude_uncurrated_cve_packages

        self.validate_mirror_filters(filters)

        if source:
            mirror["source_root"] = source

        if new_name:
            mirror["mirror_name"] = new_name

        if proxy is not None:
            mirror["proxy"] = proxy if proxy else None

        if cron is not None and cron != DEFAULT_MIRROR_CRON:
            mirror["cron"] = cron

        self.api.update_mirror(
            mirror["mirror_id"],
            channel,
            mirror["source_root"],
            mirror["mirror_name"],
            mirror["mirror_mode"],
            filters,
            mirror["mirror_type"],
            mirror["cron"],
            run_now,
            mirror["proxy"],
        )
        self.log.info("Mirror %s successfully update on channel %s", name, channel)

    def validate_mirror_filters(self, filters):
        if filters.get("only_specs") and (
            filters.get("exclude_specs") or filters.get("include_specs")
        ):
            raise errors.RepoCLIError(
                "Can't combine only_specs with exclude_specs and include_specs"
            )

    def show_list(self, channel):
        data = self.api.get_mirrors(channel)
        self.log.info(MirrorFormatter.format_list(data["items"]))
        self.log.info("")

    def show_all_list(self):
        data = self.api.get_all_mirrors()
        self.log.info(AllMirrorFormatter.format_list_all(data["items"]))
        self.log.info("")

    def sync_state(self, mirror):
        data = self.api.get_all_mirrors()
        self.log.info(SyncStateFormatter.format_sync_state(data["items"], mirror))
        self.log.info("")

    def show(self, channel, name):
        # TODO: Get mirror api is currently unavailable so we need to use the
        #       GET mirrors api and filter by name...
        # data = self.api.get_mirror(channel, name)

        mirror = self._get_by_name(channel, name)
        if mirror:
            self.log.info(MirrorFormatter.format_detail(mirror))
            self.log.info("")
        else:
            self.log.info("Mirror not found")

    def delete(self, channel, name):
        mirror = self._get_by_name(channel, name)
        if not mirror:
            self.log.info("Mirror not found")
            return
        self.api.delete_mirror(channel, mirror["mirror_id"], name)
        self.log.info("Mirror %s successfully delete on channel %s", name, channel)

    def _get_by_name(self, channel, name):
        data = self.api.get_mirrors(channel)
        for mirror in data["items"]:
            if mirror["mirror_name"] == name:
                return mirror

    def add_parser(self, subparsers):
        subparser = subparsers.add_parser(
            self.name,
            help="Manage your Anaconda repository {}s".format(self.name),
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=__doc__,
        )

        subparser.add_argument("--channel", "-c", help="Channel to mirror to.")
        group = subparser.add_mutually_exclusive_group(required=True)

        # group.add_argument('--create', nargs=2, metavar=self.name.upper())
        group.add_argument(
            "--create",
            metavar=self.name.upper(),
            help="Create a new {}".format(self.name),
        )
        group.add_argument(
            "--update", metavar=self.name.upper(), help="Update a {}".format(self.name)
        )
        group.add_argument(
            "--delete", metavar=self.name.upper(), help="Delete a {}".format(self.name)
        )
        group.add_argument(
            "--list",
            "-l",
            metavar=self.name.upper(),
            help="list all {}s for a user".format(self.name),
        )
        group.add_argument(
            "--show",
            metavar=self.name.upper(),
            help="Show all of the files in a {}".format(self.name),
        )

        group.add_argument("--list-all", action="store_true", help="list all mirrors")

        group.add_argument("--sync-state", help="show sync state of mirror")

        subparser.add_argument(
            "--source",
            "-s",
            type=check_url,
            help="Path to the source channel to mirror. "
            "I.e.: https://conda.anaconda.org/conda-test",
        )
        subparser.add_argument("--name", "-n", help="Name of the mirror")
        subparser.add_argument(
            "--mode",
            default="passive",
            help='Mirror mode. If "active", will download all the files from the source channel '
            'immediately else, if "passive", download JSON immediately and files on demand '
            'later. Default is "passive"',
        )
        subparser.add_argument(
            "--type",
            default="conda",
            help='Mirror type. Possible types: "conda", "python_simple" and "CRAN"',
        )
        subparser.add_argument(
            "--cron",
            default=DEFAULT_MIRROR_CRON,
            help="Cron string to configure the mirror job.",
        )
        subparser.add_argument(
            "--proxy",
            default=None,
            type=check_proxy_url,
            help='Proxy to use for the mirroring in format "http://<PROXYURL>" or "http://<USER>:<PASS>@<PROXYURL>"',
        )

        filters_group = subparser.add_argument_group(
            "mirror filters",
            description="Filters are used to filter specific subset of packages from the original "
            "index. Some filters are specific to artifact family.",
        )
        filters_group.add_argument(
            "--subdirs",
            default=None,
            help="[conda] List of conda subdirs, I.e.: linux-64, osx-64, linux-32, etc. "
            "Use comma-separated string.",
        )
        filters_group.add_argument(
            "--projects",
            default=None,
            help="[python] List of pypi projects to mirror. Use comma-separated string.",
        )
        filters_group.add_argument(
            "--licenses",
            default=None,
            help='List of licenses to filter for. Use comma-separated string, prepent with "-" to exclude, '
            "or just a license to include. The allowed license values are: agpl, gpl2, gpl3, lgpl, "
            "bsd, mit, apache, psf, public_domain, proprietary, other, none",
        )
        filters_group.add_argument(
            "--only_spec",
            action="append",
            default=None,
            help="MatchSpec to only spec. Use multiple times the option for multiple entries",
        )
        filters_group.add_argument(
            "--exclude_spec",
            action="append",
            default=None,
            help="MatchSpec to exclude. Use multiple times the option for multiple entries",
        )
        filters_group.add_argument(
            "--include_spec",
            action="append",
            default=None,
            help="MatchSpec to include. Use multiple times the option for multiple entries",
        )
        filters_group.add_argument(
            "--cve_score",
            type=check_cve_score,
            default=None,
            help="Only mirror files with CVE score less or equal to this value. I.e.: 9.5",
        )
        filters_group.add_argument(
            "--exclude_uncurated_cve_packages",
            default=None,
            action="store_true",
            help="Only mirror files with CVE score less or equal to this value. I.e.: 9.5",
        )
        filters_group.add_argument(
            "--date_from",
            type=check_date_format,
            default=None,
            help="[conad] bottom date YYYY-mm-dd when package was published",
        )
        filters_group.add_argument(
            "--date_to",
            type=check_date_format,
            default=None,
            help="[conad] upper date YYYY-mm-dd when package was published",
        )

        subparser.add_argument(
            "--run_now",
            action="store_true",
            help="Determines whether the mirror job should run immediately or "
            "according to the cron schedule",
        )

        subparser.set_defaults(main=self.main)
